#ifndef _E7_BIGNUM_H
#define _E7_BIGNUM_H

/* -*-C-*-

$Id: bignum.h,v 9.28 1992/09/18 16:52:34 jinx Exp $

Copyright 1986,1987,1988,1989,1992,2004 Massachusetts Institute of Technology

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

   3. The name of the author may not be used to endorse or promote
      products derived from this software without specific prior
      written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/

/* License changed to modified BSD by cph on 2004-11-08.  */

#include <string>
#include <iostream>

#include "bignum.h"


#include "config.h"


  namespace e7 {
  namespace scheme48 {

/* Internal Interface to Bignum Code */
#undef BIGNUM_ZERO_P
#undef BIGNUM_NEGATIVE_P

/* The memory model is based on the following definitions, and on the
   definition of the type `bignum_type'.  The only other special
   definition is `CHAR_BIT', which is defined in the Ansi C header
   file "limits.h". */

typedef long bignum_digit_type;
typedef long bignum_length_type;

/* BIGNUM_ALLOCATE allocates a (length + 1)-element array of
   `bignum_digit_type'; deallocation is the responsibility of the
   user (in Scheme, the garbage collector handles this). */
#define BIGNUM_ALLOCATE_TAGGED(length_in_digits)		\
    ((long *) s48_allocate_bignum((length_in_digits + 1) *	\
                                  sizeof(bignum_digit_type)))
#define BIGNUM_ALLOCATE(length_in_digits)			\
    (S48_ADDRESS_AFTER_HEADER(((char *)				\
			       BIGNUM_ALLOCATE_TAGGED((length_in_digits))), \
			       long))
extern char *	s48_allocate_bignum(long size);

/* BIGNUM_TO_POINTER casts a bignum object to a digit array pointer. */
#define BIGNUM_TO_POINTER(bignum) ((bignum_digit_type *) (bignum))

/* BIGNUM_REDUCE_LENGTH allows the memory system to reclaim some
   space when a bignum's length is reduced from its original value. */
#define BIGNUM_REDUCE_LENGTH(target, source, length)            \
     target = (long *) s48_shorten_bignum((char *) source, length)
extern char *	s48_shorten_bignum(char*, long);

/* BIGNUM_DEALLOCATE is called when disposing of bignums which are
   created as intermediate temporaries; Scheme doesn't need this. */
#define BIGNUM_DEALLOCATE(bignum) bignum_free(bignum)

/* If BIGNUM_FORCE_NEW_RESULTS is defined, all bignum-valued operations
   return freshly-allocated results.  This is useful for some kinds of
   memory deallocation strategies. */
#define BIGNUM_FORCE_NEW_RESULTS

/* BIGNUM_EXCEPTION is invoked to handle assertion violations. */
#define BIGNUM_EXCEPTION abort


#define BIGNUM_DIGIT_LENGTH (int(((sizeof (bignum_digit_type)) * CHAR_BIT) - 2))
#define BIGNUM_HALF_DIGIT_LENGTH (BIGNUM_DIGIT_LENGTH / 2)
#define BIGNUM_RADIX (((long) 1) << BIGNUM_DIGIT_LENGTH)
#define BIGNUM_RADIX_ROOT (((long) 1) << BIGNUM_HALF_DIGIT_LENGTH)
#define BIGNUM_DIGIT_MASK	 (BIGNUM_RADIX - 1)
#define BIGNUM_HALF_DIGIT_MASK	 (BIGNUM_RADIX_ROOT - 1)

#define BIGNUM_START_PTR(bignum)					\
  ((BIGNUM_TO_POINTER (bignum)) + 1)

#define BIGNUM_SET_HEADER(bignum, length, negative_p)			\
  (* (BIGNUM_TO_POINTER (bignum))) =					\
    ((length) | ((negative_p) ? BIGNUM_RADIX : 0))

#define BIGNUM_LENGTH(bignum)						\
  ((* (BIGNUM_TO_POINTER (bignum))) & ((bignum_length_type) BIGNUM_DIGIT_MASK))

#define BIGNUM_NEGATIVE_P(bignum)					\
  (((* (BIGNUM_TO_POINTER (bignum))) & BIGNUM_RADIX) != 0)

#define BIGNUM_ZERO_P(bignum)						\
  ((BIGNUM_LENGTH (bignum)) == 0)

#define BIGNUM_REF(bignum, index)					\
  (* ((BIGNUM_START_PTR (bignum)) + (index)))

#ifdef BIGNUM_FORCE_NEW_RESULTS
#define BIGNUM_MAYBE_COPY bignum_copy
#else
#define BIGNUM_MAYBE_COPY(bignum) bignum
#endif

/* These definitions are here to facilitate caching of the constants
   0, 1, and -1. */
#define BIGNUM_ZERO() S48_ADDRESS_AFTER_HEADER(s48_bignum_zero, long)
#define BIGNUM_ONE(neg_p)                                               \
   (neg_p ? S48_ADDRESS_AFTER_HEADER(s48_bignum_neg_one, long) : \
            S48_ADDRESS_AFTER_HEADER(s48_bignum_pos_one, long))

#define HD_LOW(digit) ((digit) & BIGNUM_HALF_DIGIT_MASK)
#define HD_HIGH(digit) ((digit) >> BIGNUM_HALF_DIGIT_LENGTH)
#define HD_CONS(high, low) (((high) << BIGNUM_HALF_DIGIT_LENGTH) | (low))

#define BIGNUM_BITS_TO_DIGITS(n)					\
  (((n) + (BIGNUM_DIGIT_LENGTH - 1)) / BIGNUM_DIGIT_LENGTH)

#define BIGNUM_DIGITS_FOR_LONG						\
  (BIGNUM_BITS_TO_DIGITS ((sizeof (long)) * CHAR_BIT))

#define BIGNUM_DIGITS_FOR_INT64						\
  (BIGNUM_BITS_TO_DIGITS ((sizeof (int64_t)) * CHAR_BIT))

#ifndef BIGNUM_DISABLE_ASSERTION_CHECKS

#define BIGNUM_ASSERT(expression)					\
{									\
  if (! (expression))							\
    BIGNUM_EXCEPTION ();						\
}

#endif /* not BIGNUM_DISABLE_ASSERTION_CHECKS */



/* External Interface to Bignum Code */

/* The `unsigned long' type is used for the conversion procedures
   `bignum_to_long' and `long_to_bignum'.  Older implementations of C
   don't support this type; if you have such an implementation you can
   disable these procedures using the following flag (alternatively
   you could write alternate versions that don't require this type). */
/* #define BIGNUM_NO_ULONG */

typedef long	s48_value;

#define S48_ADDRESS_AFTER_HEADER(x, type) ((type *)((x) - S48_STOB_TAG))
#define S48_STOB_TAG 3
#define S48_STOB_P(x) (((long)(x) & 3L) == S48_STOB_TAG)
#define S48_STOBTYPE_BIGNUM 20
#define S48_BIGNUM_P(x) (s48_stob_has_type(x, 20))
#define S48_STOB_ADDRESS(x) (&(S48_STOB_HEADER(x)))
#define S48_STOB_HEADER(x) (S48_STOB_REF((x),-1))
#define S48_STOB_REF(x, i) (S48_ADDRESS_AFTER_HEADER(x, s48_value)[i])


typedef void * bignum_type;
#define BIGNUM_OUT_OF_BAND ((bignum_type) 0)

enum bignum_comparison
{
  bignum_comparison_equal = 0,
  bignum_comparison_less = -1,
  bignum_comparison_greater = 1
};

typedef void * bignum_procedure_context;
extern void s48_bignum_make_cached_constants();
extern int s48_bignum_equal_p(bignum_type, bignum_type);
extern enum bignum_comparison s48_bignum_test(bignum_type);
extern enum bignum_comparison s48_bignum_compare(bignum_type, bignum_type);
extern bignum_type s48_bignum_add(bignum_type, bignum_type);
extern bignum_type s48_bignum_subtract(bignum_type, bignum_type);
extern bignum_type s48_bignum_negate(bignum_type);
extern bignum_type s48_bignum_multiply(bignum_type, bignum_type);
extern int bignum_divide(bignum_type numerator, bignum_type denominator,
			 bignum_type * quotient, bignum_type * remainder);
extern bignum_type s48_bignum_quotient(bignum_type, bignum_type);
extern bignum_type s48_bignum_remainder(bignum_type, bignum_type);
extern bignum_type s48_long_to_bignum(long);
extern bignum_type s48_ulong_to_bignum(unsigned long);
extern long s48_bignum_to_long(bignum_type);
extern unsigned long s48_bignum_to_ulong(bignum_type);
extern bignum_type s48_double_to_bignum(double);
extern double s48_bignum_to_double(bignum_type);
extern int s48_bignum_fits_in_word_p(bignum_type, long word_length,
				     int twos_complement_p);
extern bignum_type s48_bignum_length_in_bits(bignum_type);
extern bignum_type s48_bignum_length_upper_limit(void);


extern bignum_type s48_digit_stream_to_bignum
       (unsigned int n_digits,
	unsigned int (*producer)(bignum_procedure_context),
	bignum_procedure_context context,
	unsigned int radix,
	int negative_p);

extern void s48_bignum_to_digit_stream
	    (bignum_type,
	     unsigned int radix,
	     void (*consumer)(bignum_procedure_context, long),
	     bignum_procedure_context context);
extern long s48_bignum_max_digit_stream_radix(void);

/* Added bitwise operators. */

extern bignum_type s48_bignum_bitwise_not(bignum_type),
                   s48_bignum_arithmetic_shift(bignum_type, long),
                   s48_bignum_bitwise_and(bignum_type, bignum_type),
                   s48_bignum_bitwise_ior(bignum_type, bignum_type),
                   s48_bignum_bitwise_xor(bignum_type, bignum_type);

extern int s48_bignum_oddp(bignum_type);
extern long s48_bignum_bit_count(bignum_type);

extern s48_value s48_bignum_zero, s48_bignum_pos_one, s48_bignum_neg_one;

extern bignum_type
bignum_allocate_zeroed(bignum_length_type length, int negative_p);

extern void bignum_destructive_scale_up(bignum_type, bignum_digit_type);
extern void bignum_destructive_add(bignum_type, bignum_digit_type);


extern int bignum_equal_p_unsigned(bignum_type, bignum_type);
extern enum bignum_comparison bignum_compare_unsigned(bignum_type, bignum_type);

extern bignum_type bignum_add_unsigned(bignum_type, bignum_type, int);
extern bignum_type bignum_subtract_unsigned(bignum_type, bignum_type);
extern bignum_type bignum_multiply_unsigned(bignum_type, bignum_type, int);
extern bignum_type bignum_multiply_unsigned_small_factor
  (bignum_type, bignum_digit_type, int);
extern void bignum_divide_unsigned_large_denominator
  (bignum_type, bignum_type, bignum_type *, bignum_type *, int, int);
extern void bignum_destructive_normalization(bignum_type, bignum_type, int);
extern void bignum_destructive_unnormalization(bignum_type, int);
extern void bignum_divide_unsigned_normalized(bignum_type, bignum_type, bignum_type);
extern bignum_digit_type bignum_divide_subtract
  (bignum_digit_type *, bignum_digit_type *, bignum_digit_type,
   bignum_digit_type *);
extern void bignum_divide_unsigned_medium_denominator
  (bignum_type, bignum_digit_type, bignum_type *, bignum_type *, int, int);
extern bignum_digit_type bignum_digit_divide
  (bignum_digit_type, bignum_digit_type, bignum_digit_type, bignum_digit_type *);
extern bignum_digit_type bignum_digit_divide_subtract
  (bignum_digit_type, bignum_digit_type, bignum_digit_type, bignum_digit_type *);
extern void bignum_divide_unsigned_small_denominator
  (bignum_type, bignum_digit_type, bignum_type *, bignum_type *, int, int);
extern bignum_digit_type bignum_destructive_scale_down
  (bignum_type, bignum_digit_type);
extern bignum_type bignum_remainder_unsigned_small_denominator
  (bignum_type, bignum_digit_type, int);
extern bignum_type bignum_digit_to_bignum(bignum_digit_type, int);
extern bignum_type bignum_allocate(bignum_length_type, int);
extern bignum_type bignum_allocate_zeroed(bignum_length_type, int);
extern bignum_type bignum_shorten_length(bignum_type, bignum_length_type);
extern bignum_type bignum_trim(bignum_type);
extern bignum_type bignum_copy(bignum_type);
extern bignum_type bignum_new_sign(bignum_type, int);
extern bignum_type bignum_maybe_new_sign(bignum_type, int);
extern void bignum_destructive_copy(bignum_type, bignum_type);
/* Unused
extern void bignum_destructive_zero(bignum_type);
*/

/* bitwise operations */
extern bignum_type bignum_magnitude_ash(bignum_type arg1, long n);
extern bignum_type bignum_pospos_bitwise_op(int op, bignum_type, bignum_type);
extern bignum_type bignum_posneg_bitwise_op(int op, bignum_type, bignum_type);
extern bignum_type bignum_negneg_bitwise_op(int op, bignum_type, bignum_type);
extern void        bignum_negate_magnitude(bignum_type);
extern long        bignum_unsigned_logcount(bignum_type arg);
extern int         bignum_unsigned_logbitp(int shift, bignum_type bignum);

extern bignum_type bignum_read(unsigned int n_digits, const char *s, 
			       unsigned int radix, int negative_p);
extern std::string bignum_repr(bignum_type bignum, unsigned int radix);

extern void bignum_free(bignum_type b);
extern void bignum_free0(bignum_type b);
extern bignum_type bignum_maybe_copy(bignum_type b);

extern bignum_type s48_int64_to_bignum(int64_t n);
extern int64_t s48_bignum_to_int64(bignum_type bignum);

}  // namespace scheme48


class bignum_rep {
  bignum_rep &operator=(const bignum_rep &o);
  bignum_rep(const bignum_rep &o);

 public:

  size_t refcnt;

  bignum_rep *grab() {
    assert(refcnt != 0);
    refcnt++;
    return this;
  }

  void drop() {
    assert(refcnt != 0);
    --refcnt;
    if (refcnt == 0) {
      delete this;
    }
  }

  const scheme48::bignum_type b;

  bignum_rep(scheme48::bignum_type b) : refcnt(1), b(b) {
    if (b == 0) abort();
    //fprintf(stderr, "bignum_rep b = %p\n", b);
  }

  ~bignum_rep() { 
    //fprintf(stderr, "bignum_rep b = %p\n", b);
    scheme48::bignum_free(b);
  }

};

// FIXME e.g. via de-inlining
//using namespace e7::scheme48;

class Bignum {

 public:
  bignum_rep *rep;

  Bignum(const char *s, unsigned radix, int negativep) 
    : rep(new bignum_rep(scheme48::bignum_read(strlen(s), 
					       s, 
					       radix, 
					       negativep))) {
    //fprintf(stderr, "b = %p\n", b);
  }

  Bignum() : rep(new bignum_rep(scheme48::s48_int64_to_bignum(0))) {
  }
  explicit Bignum(int64_t i) 
    : rep(new bignum_rep(scheme48::s48_int64_to_bignum(i))) {
  }
  explicit Bignum(scheme48::bignum_type b) 
    : rep(new bignum_rep(b)) {
  }
  explicit Bignum(double d) 
    : rep(new bignum_rep(scheme48::s48_double_to_bignum(d))) {
  }

  Bignum(const Bignum &o) : rep(o.rep->grab()) {}

  Bignum &operator=(const Bignum &o) {
    rep->drop();
    rep = o.rep->grab();
    return *this;
  }

  ~Bignum() { rep->drop(); }

  // FIXME use static initializer?
  static void init();

  int64_t to_int64() { return scheme48::s48_bignum_to_int64(rep->b); }
  
  bool fits_int64p() { 
    return scheme48::s48_bignum_fits_in_word_p(rep->b, 64, 1);
  }

  bool zerop() {
    return scheme48::s48_bignum_test(rep->b) 
      == scheme48::bignum_comparison_equal;
  }

  bool positivep() {
    return scheme48::s48_bignum_test(rep->b) 
      == scheme48::bignum_comparison_greater;
  }

  bool negativep() {
    return scheme48::s48_bignum_test(rep->b) 
      == scheme48::bignum_comparison_less;
  }

  double to_double() {
    return scheme48::s48_bignum_to_double(rep->b);
  }

  Bignum negative() { return Bignum(scheme48::s48_bignum_negate(rep->b)); }

  Bignum operator-() { return negative(); }

  Bignum complement() { 
    return Bignum(scheme48::s48_bignum_bitwise_not(rep->b)); 
  }

  Bignum operator ~() { return complement(); }

  bool oddp() { return bool(scheme48::s48_bignum_oddp(rep->b)); }
  bool evenp() { return ! bool(scheme48::s48_bignum_oddp(rep->b)); }

  Bignum abs() { 
    if (scheme48::s48_bignum_test(rep->b) 
	== scheme48::bignum_comparison_less) {
      return Bignum(scheme48::s48_bignum_negate(rep->b));
    } else {
      return *this;
    }
  }

  // FIXME int64 version
  long bit_count() { return scheme48::s48_bignum_bit_count(rep->b); }

  // FIXME instead of above?
  Bignum length_in_bits() { 
    return Bignum(scheme48::s48_bignum_length_in_bits(rep->b)); 
  }

  bool fits_in_word(long word_length, bool twos_complementp) {
    return bool(scheme48::s48_bignum_fits_in_word_p(rep->b, word_length,
					  twos_complementp));
  }

  Bignum asl(long count) {
    return Bignum(scheme48::s48_bignum_arithmetic_shift(rep->b, count));
  }

  Bignum Bignum::length_upper_limit() { 
    return Bignum(scheme48::s48_bignum_length_upper_limit());
  }


  friend const  Bignum operator+(Bignum, Bignum);
  friend const  Bignum operator-(Bignum, Bignum);
  friend const  Bignum operator*(Bignum, Bignum);

  friend const  Bignum quotient(Bignum, Bignum);
  friend const  Bignum remainder(Bignum, Bignum);
  friend void quotient_remainder(Bignum, Bignum, Bignum *q, Bignum *r);

  friend bool operator==(Bignum, Bignum);
  friend bool operator<(Bignum, Bignum);
  friend bool operator>(Bignum, Bignum);
  friend bool operator<=(Bignum, Bignum);
  friend bool operator>=(Bignum, Bignum);
  friend bool operator!=(Bignum, Bignum);

  friend const  Bignum operator&(Bignum, Bignum);
  friend const  Bignum operator|(Bignum, Bignum);
  friend const  Bignum operator^(Bignum, Bignum);

};


extern Bignum expt(Bignum x, Bignum n);

extern Bignum gcd(Bignum a, Bignum b);
extern Bignum lcm(Bignum a, Bignum b);


std::ostream &operator<<(std::ostream &, Bignum);

}  // namespace e7


// #define E7_BIGNUM_MEM_DEBUG

#endif
